part of randomlist;

class ItemPicker {

  Completer completer;

   Random random = new Random();

   RandomList list;

   RandomListItem item;

   bool retry = false;
   int attemptCount = 1;

   Object ret = null;

   num mod = 0;

   RandomListFilter filter;

   ////////////////////////////////////////
   ItemPicker( RandomList list) {
     this.list = list;
   }



   /**
    * RETURNS ITEM, not wrapped object!
    *
     * Selects an item candidate without altering it.
     *
     * The point of allowing a 'preview' of what will be picked is to allow filtering
     * of the items based on arbitrary game constraints. E.g., only a certain class can get certain events, etc.
     *
     * can override in child to change default picking behavior.
     *
     * @param roll can pass in the 'random' number rolled in order to create a controlled testing environment or
     * a forced result.  By default the number will be completely random.
     *

    *
     *  gets fired before a result is returned from pick() to allow for filtering of list items,
     * i.e., to prevent some things from being returned by pick() based on arbitrary game logic.
     */
    Future selectCandidate([num mod=0, num roll = -1]) {
      //print("selectCandidate");
      Completer completer = new Completer();

      //print("attempt pick()");
      // by default choose random, if no # passed in
      if( roll == -1 ) roll = random.nextDouble();  // a # betw 0 and 1, inclusive

      list.lastBaseRoll = roll;

      // modifiers need to be in range 0...1
      if( mod > 0 ) {
        num adj = mod / 100;
        //trace("mod:", adj);
        roll += mod; // add in the optional modifier
      }

      list.lastModifiedRoll = roll;

      // keep within limits
      if( roll > .999999999999 ) roll = .99999999999;
      else if( roll < 0 ) roll = 0;

      //print("BASE/MOD ROLL ${list.lastBaseRoll}/, ${list.lastModifiedRoll}");

      RandomListItem candidate;

      for(RandomListItem item in list.itemList) {
          //print("  testing against item.threshold: ${item.threshold}");
          if( roll < item.threshold ) {  // pick an entry if less than threshold
              //print("...the roll fits this threshhold!");
              candidate = item;

              // ************ IF LIST ******
              // if candidate is iteself a list, pick from that one
              if( candidate.payload is RandomList ) {
                 //print("picked a child list");
                 RandomList childList = candidate.payload as RandomList;

                 // special case of an empty list, return the list itself
                 if( childList.length < 1) {
                   //print("returning empty list");
                   completer.complete(candidate);
                 } else {

                      // recalc wts for child list
                      childList.recalcWeights().then((v){
                      // attempt pick from that child list
                      childList.picker.pick(completer, returnItem:true); // NOTE: the mod and roll do NOT get passed to children
                   });
                 }
              } else {
                // return the ITEM candidate
                //print("returning candidate via completer");
                completer.complete(candidate);
              }
              break;
          } else {
              //print("...doesn't fit");
          }
      }

      return completer.future; //candidate;
    }

   bool _returnItem;

   // child list should serve up item instead of object
   pick(completer, {num mod : 0, RandomListFilter filter:null, String forceId :null, returnItem:false}) {
      attemptCount = 1;
      this.completer = completer;
      this.mod = mod;
      this.filter = filter;
      //if( filter != null) print("got filter!!!");
      _returnItem = returnItem;

      ret = null;

      try {
        // should the pick be forced?
        if( forceId != null ) {
          //print("doing a forced pick");
          // does the item exist?
          item = list.getListItem( forceId );

          if( item != null ) {
            _finalizePick( item);
            return;
          }
        }
      } catch( err ) {
        print("RandomList.pick() error during force\n" + err.message);
      }

      _tryToPick();  // attempt at least once

   } // pick

   _tryToPick() {
     //print("try to pick");
     selectCandidate( mod ).then((item){
         // item was selected

         // if there's a filter, use it
         if( filter != null ) {
            filter( item, list.pickAttemptLimit - (attemptCount++), _filterResult );
          } else {
             // otherwise finalize it right there
            _finalizePick( item );
          }
     });


   } // _try to pick



   // the callback of the filter test
   _filterResult(var item, bool isValidItem ) {
       if( isValidItem ) {
         // if it's valid
          _finalizePick( item );
       } else {
         // it was NOT valid
         // any tries left?

         if( (attemptCount <= list.pickAttemptLimit) ) {
            _tryToPick();
         } else {
             print("**** exhausted search attempt, returning null");
             _finalizePick( null );
         }
       }

   }


   /**
    * pass PAYLOAD object back through the picker's completer
    *
    */
   _finalizePick(  item) {

      //print("finalize pick ${list.id}  $item ");
       var ret;  // the object picked

       if( item != null ) {
         //trace("picked");

         list.adjustLifetime( item );
         list.itemHasBeenPicked( item );

         ret = item.payload;

         list.mostRecentItem = item;

         list.pickedCon.add(item);  // fire event
       } else {
         //trace("failed");
         //e = new Event( FAILED );
         list.failedCon.add( null );  // fire event
       }

       // if the flag is set to return the list ITEM as opposed to the PAYLOAD of the item...
       // this is the case with nested lists that have to pass the item back up
       // to the parent list
       if( _returnItem ) {
           completer.complete(item);
       } else {
           // we want to return the payload object here

           if( item.payload != null) {
               completer.complete(item.payload);
           } else {
               completer.complete(item.id);  // there is no payload so just return the ID of the item
           }
       }
    }




} // ItemPicker